#include "profiler/timer.h"
#include "aliases.h"
#include <stdio.h>
#include <string.h>
#include <time.h>

#include <x86intrin.h>

#define MAX_PROFILE_SAMPLES 1024

typedef struct {
  profiler_sample_t samples[MAX_PROFILE_SAMPLES];
  u64 cpu_freq;
  u64 size;
  u64 start;
  u64 end;
  u64 max_title_length;
} profiler_t;

INTERNAL profiler_t profiler = {0};

typedef struct timespec timespec_t;

u64 get_os_frequency() { return 1000000000; }

u64 get_os_time(void) {
  timespec_t ts = {0};

  if (clock_gettime(CLOCK_MONOTONIC_RAW, &ts) != 0) {
    return 0;
  }

  return ts.tv_sec * get_os_frequency() + ts.tv_nsec;
}

u64 read_cpu_timer(void) { return __rdtsc(); }

u64 get_cpu_freq(u64 milliseconds) {
  u64 os_freq = get_os_frequency();
  u64 os_start = get_os_time();
  u64 cpu_start = read_cpu_timer();
  u64 os_end = 0;
  u64 os_elapsed = 0;
  u64 os_wait_time = os_freq * milliseconds / 1000;

  while (os_elapsed < os_wait_time) {
    os_end = get_os_time();
    os_elapsed = os_end - os_start;
  }

  u64 cpu_end = read_cpu_timer();
  u64 cpu_elapsed = cpu_end - cpu_start;

  u64 cpu_freq = 0;

  if (os_elapsed) {
    cpu_freq = cpu_elapsed * os_freq / os_elapsed;
  }

  return cpu_freq;
}

void profile_start() {
  profiler.cpu_freq = get_cpu_freq(500);
  profiler.start = read_cpu_timer();
  profiler.max_title_length = 0;
}

void profile_end() {
  if (!profiler.start || !profiler.size) {
    return;
  }

  profiler.end = read_cpu_timer();

  u64 total = profiler.end - profiler.start;

  u16 time_precision = 16;
  u16 time_char_count = 20;
  u16 duration_char_count = 22;
  u16 percentage_precision = 8;
  u16 percentage_char_count = 12;

  // clang-format off
  printf("\n==============================PROFILING==============================\n");
  // clang-format on

  if (profiler.cpu_freq) {
    printf("Total: %*.*f seconds (CPU frequency: %llu hz/sec)\n\n",
           time_char_count, time_precision, (f64)total / profiler.cpu_freq,
           (unsigned long long)profiler.cpu_freq);
  }

  profiler_sample_t *sample = NULL;

  for (u64 i = 0; i < profiler.size; ++i) {
    sample = &(profiler.samples[i]);

    u64 duration = sample->end - sample->start;

    printf("%*s: %*lld (%*.*f %%)\n", (i32)profiler.max_title_length,
           sample->title, duration_char_count, (unsigned long long)duration,
           percentage_char_count, percentage_precision,
           (f64)duration / total * 100.0);
  }
}

profiler_sample_t sample_start(const char *title) {
  return (profiler_sample_t){
      .title = title,
      .start = read_cpu_timer(),
      .end = 0,
  };
}

void sample_end(profiler_sample_t *sample) {
  if (profiler.size == MAX_PROFILE_SAMPLES) {
    return;
  }

  u64 length = strlen(sample->title);

  if (length > profiler.max_title_length) {
    profiler.max_title_length = length;
  }

  sample->end = read_cpu_timer();

  profiler.samples[(profiler.size)++] = *sample;
}
